/* -----------------------------------------------------------------------

The MIT License (MIT)

Copyright (c) 2014 Huayan Wang

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

If you use this software, please cite [1]. If you use it with cycles turned on (default),
please cite [1] and [2].

[1] Huayan Wang and Daphne Koller: Subproblem-Tree Calibration: A Unified Approach to Max-Product Message Passing, ICML 2013
[2] Huayan Wang and Daphne Koller: A Fast and Exact Energy Minimization Algorithm for Cycle MRFs, ICML 2013

----------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include "dualbca.h"

void showProblem(MapProblem* problem)
{
  printf("-------------------------\n");
  printf("problem->num_variable = %d\n", problem->num_variable);
  printf("problem->num_subproblem = %d\n", problem->num_subproblem);
  for(int i=0; i<problem->num_subproblem; ++i){
    SubProblem *p = problem->sp[i];
    printf("spid%d : ( ", i);
    for(int j=0; j<p->num_variable; ++j)
      printf("%d ", p->var_id[j]);
    printf(") ");
    printf("phiofsp : ( ");
    for(int j=0; j<problem->num_phi_of_sp[i]; ++j)
      printf("%d ", problem->phi_of_sp[i][j]);
    printf(")\n");
  }

  printf("problem->num_phi = %d\n", problem->num_phi);
  for(int i=0; i<problem->num_phi; ++i){
    if(i < problem->num_variable)
      printf("phiid%d,  (%d) (", i, problem->scope_of_phi[i][0]);
    else
      printf("phiid%d,  (%d, %d) (", i, problem->scope_of_phi[i][0], problem->scope_of_phi[i][1]);

    for(int j=0; j<problem->num_sp_of_phi[i]; ++j)
      printf("%d ", problem->sp_of_phi[i][j]);
    printf(")\n");
  }
  printf("-------------------------\n");
}


int isSparseGraph(MapProblem* problem)
{
  int num_edges = 0;
  int magic_num = 8*(problem->num_variable); 
  for(int i = 0; i<problem->num_subproblem; ++i){
    SubProblem* p = problem->sp[i];
    if(p->num_variable > 1){
      num_edges += (p->num_variable * (p->num_variable-1));
    }
  }
  if(num_edges < magic_num)
    return 1;
  else
    return 0;
}

void writeSolutionFile(const char* outputfile, int N, int* sol)
{
  FILE* fp = fopen(outputfile, "w");
  fprintf(fp, "%d ", N);
  for(int i=0; i<N; ++i)
    fprintf(fp, "%d ", sol[i]);
  fprintf(fp, "\n");
  fclose(fp);
}

void removeFromList(int val, int* vec, int n)
{
  int i=0;
  while(vec[i]!=val && i<n)
    ++i;
  if(i<n-1)
    vec[i] = vec[n-1];
}


void incorporateEvidence(MapProblem* problem, MapParameters* parameters, const char* filename)
{
  FILE* fp = fopen(filename, "r");
  if(!fp) return;
  int num_evid=0;
  fscanf(fp, "%d", &num_evid);
  for(int evid=0; evid<num_evid; ++evid){
    int var;
    int assignment;
    fscanf(fp, "%d %d", &var, &assignment);

    for(int i=0; i<problem->num_sp_of_phi[var]; ++i){
      int sp_id = problem->sp_of_phi[var][i];
      SubProblem* p = problem->sp[sp_id];
      Function* f = &(problem->func[p->func_id]);
      int i0 = findInVec(var, p->var_id, p->num_variable);

      int N = p->num_variable;
      int* stride = p->sol;
      p->min_energy_valid = 0;
      stride[0] = 1;
      for(int j=1; j<N; ++j)
	stride[j] = stride[j-1] * (p->card[j-1]);
      assert(stride[N-1] * p->card[N-1] == f->size_val);
      for(int index=0; index<(f->size_val); ++index)
	if((index / stride[i0]) % (p->card[i0]) != assignment)
	  p->val[index] = INFVAL;
      removeFromList(var, problem->phi_of_sp[sp_id], problem->num_phi_of_sp[sp_id]);
      problem->num_phi_of_sp[sp_id]--;
    }
    problem->num_sp_of_phi[var] = 0; // never need to pass msg over this scope again
    problem->sol[var] = assignment;
  }
  fclose(fp);
  problem->primal_obj = 1e8;
  renewMRFSolution(problem, parameters);
  problem->dual_obj = evaluateDual(problem, parameters);
  problem->init_dual_obj = problem->dual_obj;
  problem->init_primal_obj = problem->primal_obj;
  problem->last_dual_gap = problem->primal_obj - problem->dual_obj;
}


int main(int argc, char* argv[])
{

  if(argc < 5 || argc > 7){
    printf("Usage:\n");
    printf("./solver <input-model-file> <input-evidence-file> <output-file> <time-budget-in-seconds> [options]\n");
    exit(1);
  }

  MapProblem *problem;
  MapParameters parameters;

  int verbose = 1;

  clock_t tic = clock();

  if(strcmp(getFilenameExt(argv[1]), "map")==0)
    readProblemTXT(&problem, argv[1]);
  else if(strcmp(getFilenameExt(argv[1]), "uai")==0)
    readProblemUAI(&problem, argv[1]);
  else 
    {printf("ERR: incorrect input file extension\n"); exit(1);}

  readParameters(&parameters, argc-5, argv+5);

  if(verbose && parameters.add_cycle_batchsize == 0) 
    printf("warning: cycles are turned off, turn it on for best performance\n");

  if(isSparseGraph(problem)){
    if(verbose) printf("using sparse graph method\n");
    parameters.tree_selection = TREE_SP2;
  }else{
    if(verbose) printf("using dense graph method\n");
    parameters.tree_selection = TREE_STARP;
  }

  int terminate = 0;
  double timebudget = atof(argv[4]);
  if(verbose) printf("time budget = %.2f seconds\n", timebudget);

  double time_per_iter = 0.0;
  double total_time = 0.0;

  initMapProblem(problem, &parameters);
  incorporateEvidence(problem, &parameters, argv[2]);

  // showProblem(problem);

  while(terminate==0 && total_time < timebudget){

    clock_t tic0 = clock();
    terminate = runInference(problem, &parameters);
    time_per_iter = ((double)clock()-tic0)/CLOCKS_PER_SEC;

    // showProblem(problem);

    total_time = ((double)clock()-tic)/CLOCKS_PER_SEC;
    if(verbose)
      printf("%.2f secs, energy=%f, bound=%f, progress=%f, cycles=%d\n", 
	     total_time,
	     problem->primal_obj, 
	     problem->dual_obj,
	     problem->dual_gap_decrease,
	     problem->used_cycles);

    if(total_time + 3*time_per_iter > timebudget)
      writeSolutionFile(argv[3], problem->num_variable, problem->sol);
  }
  writeSolutionFile(argv[3], problem->num_variable, problem->sol);
  freeMapProblem(&problem);
  return 0;
}
